/*
 * Copyright 2012, 2017 INFN
 *
 * Licensed under the EUPL, Version 1.2 or – as soon they
 * will be approved by the European Commission - subsequent
 * versions of the EUPL (the "Licence");
 * You may not use this work except in compliance with the
 * Licence.
 * You may obtain a copy of the Licence at:
 *
 * https://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in
 * writing, software distributed under the Licence is
 * distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 * express or implied.
 * See the Licence for the specific language governing
 * permissions and limitations under the Licence.
 */

#include <chaos/common/batch_command/BatchCommand.h>

#include <chaos/cu_toolkit/control_manager/SCAbstractControlUnit.h>

#include <chaos/cu_toolkit/control_manager/slow_command/command/SCWaitCommand.h>
#include "../windowsCompliant.h"

using namespace chaos;

using namespace chaos::cu;
using namespace chaos::cu::control_manager;
using namespace chaos::cu::control_manager::slow_command;
using namespace chaos::cu::control_manager::slow_command::command;
using namespace chaos::common::data;
using namespace chaos::common::data::cache;
using namespace chaos::common::utility;
using namespace chaos::common::batch_command;

#define SCACU_LAPP_	INFO_LOG(SCAbstractControlUnit)
#define SCACU_LDBG_	DBG_LOG(SCAbstractControlUnit)
#define SCACU_LERR_	ERR_LOG(SCAbstractControlUnit)

SCAbstractControlUnit::SCAbstractControlUnit(const std::string& _control_unit_id,
                                             const std::string& _control_unit_param)
    : AbstractControlUnit(CUType::SCCU,
                          _control_unit_id,
                          _control_unit_param)
    , slow_command_executor(NULL) {
  //allocate the SlowCommandExecutor, custom implementation of batch command executor engine
  slow_command_executor = new SlowCommandExecutor(control_unit_instance, this);
  //set the driver delegator to the executor
  slow_command_executor->driverAccessorsErogator = this;
  //associate the shared cache of the executor to the asbtract control unit one
  attribute_value_shared_cache = slow_command_executor->getAttributeSharedCache();
}

/*!
 Parametrized constructor
 \param _control_unit_id unique id for the control unit
 \param _control_unit_drivers driver information
 */
SCAbstractControlUnit::SCAbstractControlUnit(const std::string&           _control_unit_id,
                                             const std::string&           _control_unit_param,
                                             const ControlUnitDriverList& _control_unit_drivers)
    : AbstractControlUnit(CUType::SCCU,
                          _control_unit_id,
                          _control_unit_param,
                          _control_unit_drivers)
    , slow_command_executor(NULL) {
  //allocate the SlowCommandExecutor, custom implementation of batch command executor engine
  slow_command_executor = new SlowCommandExecutor(control_unit_instance, this);
  //set the driver delegator to the executor
  slow_command_executor->driverAccessorsErogator = this;
  //associate the shared cache of the executor to the asbtract control unit one
  attribute_value_shared_cache = slow_command_executor->getAttributeSharedCache();
}

SCAbstractControlUnit::~SCAbstractControlUnit() {
  if (slow_command_executor) {
    delete (slow_command_executor);
    slow_command_executor = NULL;
  }
}

void SCAbstractControlUnit::_getDeclareActionInstance(std::vector<const DeclareAction*>& declareActionInstance) {
  //broadcast to the parent the method call
  AbstractControlUnit::_getDeclareActionInstance(declareActionInstance);

  //fill the vecto rwith the local declare action instance
  declareActionInstance.push_back(slow_command_executor);
}

//! called whr the infrastructure need to know how is composed the control unit
void SCAbstractControlUnit::_defineActionAndDataset(CDataWrapper& setup_configuration) {
  //add the batch command description to the configuration
  SCACU_LDBG_ << "Install default slow commands for node id:" << DatasetDB::getDeviceID();
  installCommand(BATCH_COMMAND_GET_DESCRIPTION(SCWaitCommand), false);

  //call superclass method
  AbstractControlUnit::_defineActionAndDataset(setup_configuration);
  addStateVariable(StateVariableTypeAlarmCU, "command_fault", "Notify when a command faulted");

  std::vector<ChaosSharedPtr<BatchCommandDescription> > batch_command_description;
  slow_command_executor->getCommandsDescriptions(batch_command_description);
  if (batch_command_description.size()) {
    //fill setup with command descirption serialization
    for (std::vector<ChaosSharedPtr<BatchCommandDescription> >::iterator it = batch_command_description.begin();
         it != batch_command_description.end();
         it++) {
      ChaosSharedPtr<CDataWrapper> full_description((*it)->getFullDescription());
      setup_configuration.appendCDataWrapperToArray(*full_description);
    }
    setup_configuration.finalizeArrayForKey(chaos::ControlUnitNodeDefinitionKey::CONTROL_UNIT_DATASET_COMMAND_DESCRIPTION);
  }
}

AbstractSharedDomainCache* SCAbstractControlUnit::_getAttributeCache() {
  return AbstractControlUnit::_getAttributeCache();
}

/*
 Initialize the Custom Contro Unit and return the configuration
 */
void SCAbstractControlUnit::init(void* initData) {
  //this need to be made first of AbstractControlUnit::init(initData); because
  //SlowCommandExecutor has his own instance of cache and this control unit need to
  //use that

  //call parent impl
  AbstractControlUnit::init(initData);

  //control unit is it'self the database
  slow_command_executor->control_unit_instance = this;

  //init executor
  StartableService::initImplementation(slow_command_executor, (void*)NULL, "Slow Command Executor", __PRETTY_FUNCTION__);

  SCACU_LDBG_ << "Initializing slow command sandbox for node id " << DatasetDB::getDeviceID();

  //now we can call funciton for custom definition of the shared variable
  SCACU_LDBG_ << "Setting up custom shared variable for node id " << DatasetDB::getDeviceID();
}

/*
 Deinit the Control Unit
 */
void SCAbstractControlUnit::deinit() {
  //call parent impl

  if (slow_command_executor) {
    SCACU_LDBG_ << "Deinitialize the command executor for " << DatasetDB::getDeviceID();
    StartableService::deinitImplementation(slow_command_executor, "Slow Command Executor", __PRETTY_FUNCTION__);
    //deassociate the data storage
    slow_command_executor->control_unit_instance = NULL;
  } else {
    SCACU_LDBG_ << "No command executor allocated for " << DatasetDB::getDeviceID();
  }
  AbstractControlUnit::deinit();
}

/*
 Starto the  Control Unit scheduling for node
 */
void SCAbstractControlUnit::start() {
  //call parent impl
  AbstractControlUnit::start();

  SCACU_LDBG_ << "Start sandbox for node id:" << DatasetDB::getDeviceID();
  StartableService::startImplementation(slow_command_executor, "Slow Command Executor", __PRETTY_FUNCTION__);
}

/*
 Stop the Custom Control Unit scheduling for node
 */
void SCAbstractControlUnit::stop() {
  //call parent impl
  AbstractControlUnit::stop();

  SCACU_LDBG_ << "Stop slow command executor for node id:" << DatasetDB::getDeviceID();
  StartableService::stopImplementation(slow_command_executor, "Slow Command Executor", __PRETTY_FUNCTION__);
}

// Perform a command registration
void SCAbstractControlUnit::setDefaultCommand(const std::string& dafaultCommandName,
                                              bool               sticky,
                                              unsigned int       sandbox_instance) {
  slow_command_executor->setDefaultCommand(dafaultCommandName,
                                           sandbox_instance);
}

void SCAbstractControlUnit::addExecutionChannels(unsigned int execution_channels) {
  slow_command_executor->addSandboxInstance(execution_channels);
}

//! system dataset configuraiton overload
void SCAbstractControlUnit::initSystemAttributeOnSharedAttributeCache() {
  //first call root system dataset initialization method
  AbstractControlUnit::initSystemAttributeOnSharedAttributeCache();

  //add slow control system prorperty
  AttributeCache& domain_attribute_setting = attribute_value_shared_cache->getSharedDomain(DOMAIN_SYSTEM);
  //add heart beat attribute
  SCACU_LDBG_ << "Adding additional system attribute for control unit numbers";
  domain_attribute_setting.addAttribute(DataPackSystemKey::DP_SYS_RUN_UNIT_AVAILABLE, 0, DataType::TYPE_INT32);

  uint32_t run_unit = slow_command_executor->getNumberOfSandboxInstance();
  domain_attribute_setting.setValueForAttribute(domain_attribute_setting.getNumberOfAttributes() - 1, &run_unit, sizeof(uint32_t));

  std::vector<std::string> all_sandbox_ids;
  slow_command_executor->getSandboxID(all_sandbox_ids);

  //add sand box identifier for the last system event
  domain_attribute_setting.addAttribute(DataPackSystemKey::DP_SYS_RUN_UNIT_ID, 127, DataType::TYPE_STRING);

  
}

void SCAbstractControlUnit::completeInputAttribute() {
}

void SCAbstractControlUnit::submitSlowCommand(const std::string command_alias,
                                              CDataWrapper*     slow_command_pack,
                                              uint64_t&         command_id) {
  CHAOS_ASSERT(slow_command_executor)
  slow_command_executor->submitCommand(command_alias,
                                       slow_command_pack,
                                       command_id);
}
void SCAbstractControlUnit::consumerHandler(const chaos::common::message::ele_t& data){
      std::string rootname = data.key;
      size_t      pos      = rootname.find(DataPackPrefixID::COMMAND_DATASET_POSTFIX);
      if (pos != std::string::npos) {
        //command queue
        uint64_t         command_id;
        if(data.cd->hasKey(chaos_batch::BatchCommandAndParameterDescriptionkey::BC_ALIAS)){
          std::string bc_alias=data.cd->getStringValue(chaos_batch::BatchCommandAndParameterDescriptionkey::BC_ALIAS);
          
          SCACU_LDBG_ << "Submitting command:"<<bc_alias<<" data:"<<data.cd->getJSONString();

          slow_command_executor->submitCommand(bc_alias,
                                       data.cd->clone().release(),
                                       command_id);
          }
          return;
      }
     SCAbstractControlUnit::consumerHandler(data);
}

void SCAbstractControlUnit::submitBatchCommand(const std::string&        batch_command_alias,
                                               chaos_data::CDataWrapper* command_data,
                                               uint64_t&                 command_id,
                                               uint32_t                  execution_channel,
                                               uint32_t                  priority,
                                               uint32_t                  submission_rule,
                                               uint32_t                  submission_retry_delay,
                                               uint64_t                  scheduler_step_delay) {
  CHAOS_ASSERT(slow_command_executor)
  slow_command_executor->submitCommand(batch_command_alias,
                                       command_data,
                                       command_id,
                                       execution_channel,
                                       priority,
                                       submission_rule,
                                       submission_retry_delay,
                                       scheduler_step_delay);
}

ChaosUniquePtr<CommandState> SCAbstractControlUnit::getStateForCommandID(uint64_t command_id) {
  return slow_command_executor->getStateForCommandID(command_id);
}

CDWUniquePtr SCAbstractControlUnit::setDatasetAttribute(CDWUniquePtr dataset_attribute_values) {
  uint64_t                                          command_id = 0;
  ChaosUniquePtr<chaos::common::data::CDataWrapper> result_for_command;

  //cal first the superclass method because the dataset_attribute_values is not detached
  CDWUniquePtr result;

  //check if we have a command
  if (dataset_attribute_values->hasKey(chaos_batch::BatchCommandAndParameterDescriptionkey::BC_ALIAS)) {
    CHAOS_ASSERT(slow_command_executor)
    std::string command_alias = dataset_attribute_values->getStringValue(chaos_batch::BatchCommandAndParameterDescriptionkey::BC_ALIAS);
    // in slow control cu the CDataWrapper instance received from rpc is internally managed
    //so we need to detach it
    // submit the detacched command to slow controll subsystem
    slow_command_executor->submitCommand(command_alias,
                                         dataset_attribute_values.release(),
                                         command_id);
    //construct the result if we don't already have it
    if (!result.get()) {
      result.reset(new CDataWrapper());
    }
    //add command id into the result
    result->addInt64Value(chaos_batch::BatchCommandExecutorRpcActionKey::RPC_GET_COMMAND_STATE_CMD_ID_UI64, command_id);
  } else {
    // or command or setAttribute not both 
    result= AbstractControlUnit::setDatasetAttribute(MOVE(dataset_attribute_values->clone()));
  }
  return result;
}

void SCAbstractControlUnit::propertyUpdatedHandler(const std::string&  group_name,
                                                   const std::string&  property_name,
                                                   const CDataVariant& old_value,
                                                   const CDataVariant& new_value) {
  if (group_name.compare(chaos::ControlUnitPropertyKey::P_GROUP_NAME) == 0) {
    key_data_storage->updateConfiguration(property_name, new_value);
    //is my group
    if (property_name.compare(ControlUnitDatapackSystemKey::THREAD_SCHEDULE_DELAY) == 0) {
      chaos_batch::features::Features features;
      std::memset(&features, 0, sizeof(chaos_batch::features::Features));
      features.defaultSchedulerStepsDelay=features.featureSchedulerStepsDelay = new_value.asUInt64();

      slow_command_executor->setCommandFeatures(features);
      _updateRunScheduleDelay(new_value.asUInt64());
      //pushSystemDataset();
    }
  }
  //call superclass
  AbstractControlUnit::propertyUpdatedHandler(group_name,
                                              property_name,
                                              old_value,
                                              new_value);
}

void SCAbstractControlUnit::installCommand(ChaosSharedPtr<BatchCommandDescription> command_description,
                                           bool                                    is_default,
                                           bool                                    sticky,
                                           bool                                    auto_busy,
                                           unsigned int                            sandbox) {
  CHAOS_ASSERT(slow_command_executor)
  //set custom attribute
  bool loc_auto_busy = auto_busy;
  slow_command_executor->installCommand(command_description);
  if (is_default) {
    //default command usually doesn't need to show busy
    loc_auto_busy = false;
    setDefaultCommand(command_description->getAlias(),
                      sticky,
                      sandbox);
  }
  command_description->getCustomAttributeRef().insert(BCInstantiationAttributeMapPair("auto_busy", CDataVariant(loc_auto_busy)));
}
bool SCAbstractControlUnit::waitOnCommandID(uint64_t& cmd_id) {
  if(getState()!=chaos::CUStateKey::START){
		return true;
	}
  ChaosUniquePtr<CommandState> cmd_state;
  do {
    cmd_state = getStateForCommandID(cmd_id);
    if (!cmd_state.get()) break;

    switch (cmd_state->last_event) {
      case BatchCommandEventType::EVT_QUEUED:
        SCACU_LDBG_ << cmd_id << " -> QUEUED";
        break;
      case BatchCommandEventType::EVT_RUNNING:
        SCACU_LDBG_ << cmd_id << " -> RUNNING";
        break;
      case BatchCommandEventType::EVT_WAITING:
        SCACU_LDBG_ << cmd_id << " -> WAITING";
        break;
      case BatchCommandEventType::EVT_PAUSED:
        SCACU_LDBG_ << cmd_id << " -> PAUSED";
        break;
      case BatchCommandEventType::EVT_KILLED:
        SCACU_LDBG_ << cmd_id << " -> KILLED";
        break;
      case BatchCommandEventType::EVT_COMPLETED:
        SCACU_LDBG_ << cmd_id << " -> COMPLETED";
        break;
      case BatchCommandEventType::EVT_FAULT:
        SCACU_LDBG_ << cmd_id << " -> FAULT";
        break;
      default:
        break;
    }
    //whait some times
    usleep(500000);
  } while (cmd_state->last_event != BatchCommandEventType::EVT_COMPLETED &&
           cmd_state->last_event != BatchCommandEventType::EVT_FAULT &&
           cmd_state->last_event != BatchCommandEventType::EVT_KILLED);
  return (cmd_state.get() &&
          cmd_state->last_event == BatchCommandEventType::EVT_COMPLETED);
}
